"""Module defining the `JoyStickEvent` class (NOT IMPLEMENTED)."""

from ..event import Event


# TODO joystick support may be provided at some point
class JoyStickEvent(Event):
    """A class representing a joystick event (NOT IMPLEMENTED)."""

    pass
