"""Package defines UI related functionality. This functionality may be used by `star_ray` extensions to form a common public API."""

from .window_configuration import WindowConfiguration

__all__ = ("WindowConfiguration",)
